/*
Very simple GLUT program using OSL libraries to
demonstrate use of GL_ARB_shading_language_100.

Orion Sky Lawlor, olawlor@acm.org, 2005/2/10 (Public Domain)
*/
#include "ogl/ext.h"
#include "ogl/main.h"
#include "osl/osl.h"
#include "osl/viewpoint.h" /* 3D viewpoint utility routines */
#include <GL/glut.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <vector>

/* Include library implementations here for fast linking */
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"
#include "ogl/ext.cpp"
#include "ogl/glew.c"
#include "ogl/main.cpp"
#include "ogl/readback.cpp"
#include "ogl/program.cpp"
#include "osl/osl.cpp"
#include "osl/viewpoint.cpp"
void osl::bad(const char *a,const char *b,const char *c,int d) 
{ printf(a,b,c,d); exit(1);}

#include "frag_bench.cpp"

/************ Main Drawing Function **************/

/**
  Program to run on each vertex or pixel fragment.
*/
class oglShaderObject {
	int target;// Program target, like GL_FRAGMENT_SHADER_ARB
	GLhandleARB progName; // OpenGL handle to program
public:
	oglShaderObject(int target_) :target(target_) 
		{progName=glCreateShaderObjectARB(target);}
	
	/// Get our handle (e.g., for glAttachObjectARB)
	GLhandleARB get(void) const {return progName;}
	
	// Replace the current program by the one in this file.
	//  Prints out an error message if the program can't be parsed.
	void read(const char *fName);
	
	// Replace the current program by this one.
	//  Prints out an error message if the program can't be parsed.
	void set(const char *programText);
	
	~oglShaderObject() {glDeleteObjectARB(progName);}
};

void oglShaderObject::read(const char *fName) {
	
	// Read program from file
	int len=65536;
	char *s=new char[len];
	FILE *f=fopen(fName,"r");
	len=fread(s,1,len,f); // reset len to file length
	s[len]=0; // null terminate string
	fclose(f);
	
	// Upload program to graphics card
	set(s);
	
	delete[] s;
}

void checkOp(GLhandleARB obj,int what,const char *where) {
	GLint compiled;
	glGetObjectParameterivARB(obj,what,&compiled);
	if (!compiled) {
		printf("Compile error on program: %s\n",where);
		enum {logSize=1000};
		char log[logSize]; int len=0;
		glGetInfoLogARB(obj, logSize,&len,log);
		printf("Log: \n%s\n",log);
		exit(1);
	}
}
	
void oglShaderObject::set(const char *programText) {
	glShaderSourceARB(progName,1, &programText,NULL);
	glCompileShaderARB(progName);
	checkOp(progName,GL_OBJECT_COMPILE_STATUS_ARB,programText);
	oglCheck(programText);
}
void uploadShaders(oglShaderObject &vert,oglShaderObject &frag,
	GLhandleARB &prog,const oglViewpoint &vp) 
{
	double start_upload=osl::time();
	
	vert.read("vertex.txt");
	frag.read("fragment.txt");
	
	glAttachObjectARB(prog,vert.get());
	glAttachObjectARB(prog,frag.get());

	glLinkProgramARB(prog);
	checkOp(prog,GL_OBJECT_LINK_STATUS_ARB,"link");
	oglCheck("link");
	
	if (0) { /* Linker log can show what runs on hardware */
		char log[65536];int loglen=0;
		glGetInfoLogARB(prog,sizeof(log)-1,&loglen,log);
		log[loglen]=0;
		printf("Log: %s\n",log);
	}
	
	glUseProgramObjectARB(prog);
	oglCheck("use");
	
	/* copy camera location into vertex shader attribute */
	oglVector3d cameraLocation=vp.getEye();
	glUniform4fARB(glGetUniformLocationARB(prog,"cameraLocation"),
		cameraLocation.x,cameraLocation.y,cameraLocation.z,1);
	
	/* Set up texture "sampler" as texture unit 0 */
	glUniform1iARB(glGetUniformLocationARB(prog,"myNoise"),0);
	
	double elapsed_upload=osl::time()-start_upload;
	if (oglToggles['t'])
		printf("Upload: %.2f ms\n",1.0e3*elapsed_upload);
}


class myVertex {
public:
	osl::Vector3f loc; /* Location of vertex */
	myVertex(const osl::Vector3f &l) :loc(l) {}
};

class fragBench : public oglDrawable {
public:
	std::vector<myVertex> vertices;
	GLuint vertex_buffer;
	
	fragBench();
	virtual void draw(const oglViewpoint &vp);
};

/******** Program Initialization *********/
int main(int argc, char** argv)
{
	oglOptions i;
	i.name="GL_ARB_shader_language_100 Test";
	i.clearColor=oglVector3d(0.2,0.3,0.4);
	oglSetup(i);
	oglSetupExt();
	if (!GLEW_ARB_fragment_shader) {
		printf("You need programmable shading to run this program!\n");
		exit(1);
	}
	uploadNoise(8);

	oglEventloop(new oglTrackballController(),new fragBench());
	
	return 0;
}

/* Silly little height function */
osl::Vector3f makeVertex(double x,double y) {
	double z=
		0.10*cos(125*(x*x+y*y))+
		0.05*cos(137*((x+0.1)*(x+0.1)+y*y));
	return osl::Vector3f(x,y,z);
}

fragBench::fragBench()
{
	int n=33; 
	if (1) n=1000; /* 1000 x 1000 array! */
	double sz=1.0/n;
	for (int iy=0;iy<n;iy++) {
		double x=0*sz-0.5,y=iy*sz-0.5;
		vertices.push_back(makeVertex(x,y));
		for (int ix=0;ix<n;ix++) {
			x=ix*sz-0.5;
			vertices.push_back(makeVertex(x,y));
			vertices.push_back(makeVertex(x,y+sz));
		}
		vertices.push_back(makeVertex(x,y+sz));
		
	}
	vertex_buffer=0;
	glGenBuffersARB(1, &vertex_buffer);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,
		 sizeof(myVertex)*vertices.size(), 
		 &vertices[0], 1?GL_STATIC_DRAW_ARB:GL_STREAM_DRAW_ARB);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
}

void fragBench::draw(const oglViewpoint &vp)
{
	oglShaderObject vert(GL_VERTEX_SHADER_ARB);   
	oglShaderObject frag(GL_FRAGMENT_SHADER_ARB); 
	
	GLhandleARB prog=glCreateProgramObjectARB();
	if (!oglToggles['f']) uploadShaders(vert,frag,prog,vp);
	
	glDisable(GL_BLEND);
	
	if (!oglToggles['a']) 
	{ /* Array of n x n triangles */
		
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertex_buffer);
		glEnableClientState(GL_VERTEX_ARRAY);	
		glVertexPointer(3,GL_FLOAT,
			sizeof(myVertex),0);
		glDrawArrays(GL_TRIANGLE_STRIP, 0, vertices.size());
		
		glDisableClientState(GL_VERTEX_ARRAY);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
		
		oglRepaint(10);
	}
	
	glUseProgramObjectARB(0);
	glDeleteObjectARB(prog);
}
